#!/usr/bin/env python3

import sys
import os
import stat
import argparse
import base64

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apcommon import apsecretclient


class GenevaSecretRefresher:
    SECRET_MDSD_CERT = "Geneva-GCS-Certificate"
    SECRET_MDSD_KEY = "Geneva-GCS-Key"

    SECRET_MDSDINT_CERT = "GenevaINT-GCS-Certificate"
    SECRET_MDSDINT_KEY = "GenevaINT-GCS-Key"

    GCS_ENVIRONMENT = "DiagnosticsProd"
    GCS_ENVIRONMENT_INT = "Test"

    GCS_ACCOUNT = "APLinux"

    GCS_ENV_FILENAME = "mdsd_vars.env"
    GCS_CERT_FILENAME = "gcscert.pem"
    GCS_KEY_FILENAME = "gcskey.pem"

    APSECRET_ENDPOINTS = ["apsecret.aplinuxmon-prod-co4.co4.ap.gbl:1443",

                          # Legacy APSecret servers
                          # "aplinuxservices.trafficmanager.net:1443"
                          ]

    def __init__(self, client_key_file, client_cert_file):
        self.mdsd_environment = {}
        self.mdsd_environment["azlinux-dev-co4"] = "int"
        self.mdsd_environment["azlinux-imagebuild-cy2test01"] = "int"
        self.mdsd_environment["azlinux-test-co4"] = "int"
        self.mdsd_environment["azlinux-test-cy2test01"] = "int"
        self.mdsd_environment["charlilieap-dev-co3"] = "int"
        self.mdsd_environment["charlilieap-dev-cy2test01"] = "int"
        self.mdsd_environment["charlilieap-dev-cy2test02"] = "int"
        self.mdsd_environment["charlililinux-dev-cy2test02"] = "int"
        self.mdsd_environment["charlililinux1-dev-co3c"] = "int"

        self.default_mdsd_environment = "prod"

        self.key_file = client_key_file
        self.cert_file = client_cert_file
        self.kv_client = None

        self.mdsd_cert = ""
        self.mdsd_key = ""

        # DC/Region Mappings for Geneva Warm Path Autokey Zones
        # See http://sharepoint/sites/autopilot/wiki/Virtual%20Machines.aspx
        # See also https://msazure.visualstudio.com/AzureWiki/_wiki/wikis/AzureWiki.wiki?pagePath=%2FAzure%20Wiki%20Home%2FAutopilot%2FEAP%2FEAP%20Regions&wikiVersion=GBwikiMaster
        self.gcs_regions = {}
        self.gcs_regions["BJ1"] = "eastasia"  # china
        self.gcs_regions["BN1"] = "eastus"  # useast
        self.gcs_regions["BN2"] = "eastus"  # useast
        self.gcs_regions["BN2B"] = "eastus"  # useast
        self.gcs_regions["BN4C"] = "eastus"  # useast
        self.gcs_regions["CH01"] = "northcentralus"  # usnorth
        self.gcs_regions["CH1"] = "northcentralus"  # usnorth
        self.gcs_regions["CH1B"] = "northcentralus"  # usnorth
        self.gcs_regions["CH1D"] = "northcentralus"  # usnorth
        self.gcs_regions["CHI02"] = "northcentralus"  # usnorth
        self.gcs_regions["CO01"] = "westus2"  # uswest
        self.gcs_regions["CO1C"] = "westus2"  # uswest
        self.gcs_regions["CO3"] = "westus2"  # uswest
        self.gcs_regions["CO3B"] = "westus2"  # uswest
        self.gcs_regions["CO3C"] = "westus2"  # uswest
        self.gcs_regions["CO4"] = "westus2"  # uswest
        self.gcs_regions["CO4C"] = "westus2"  # uswest
        self.gcs_regions["CY2"] = "westcentralus"  # uscentral
        self.gcs_regions["CY2B"] = "westcentralus"  # uscentral
        self.gcs_regions["CY2TEST01"] = "westcentralus"  # uscentral
        self.gcs_regions["CY2TEST02"] = "westcentralus"  # uscentral
        self.gcs_regions["CY4"] = "westcentralus"  # uscentral
        self.gcs_regions["DB3"] = "northeurope"  # europe
        self.gcs_regions["DB4"] = "northeurope"  # europe
        self.gcs_regions["DB5"] = "northeurope"  # europe
        self.gcs_regions["DM2C"] = "centralus"  # uscentral
        self.gcs_regions["DM3"] = "centralus"  # uscentral
        self.gcs_regions["DUB01"] = "northeurope"  # europe
        self.gcs_regions["DUB02"] = "northeurope"  # europe
        self.gcs_regions["HK2"] = "eastasia"  # asia
        self.gcs_regions["HKG01"] = "eastasia"  # asia
        self.gcs_regions["MW1"] = "westus2"  # uswest
        self.gcs_regions["MWH01"] = "westus2"  # uswest
        self.gcs_regions["SG1"] = "southeastasia"  # asia
        self.gcs_regions["SN2"] = "southcentralus"  # ussouth

        self.gcs_default_region = "westus2"

        self.mdsd_gcs_environment = ""
        self.mdsd_gcs_region = ""

    def __is_connected(self) -> bool:
        if self.kv_client is None:
            return False

        # TODO: Make this more resilient
        return True

    def __connect(self):
        if self.__is_connected():
            return
        self.kv_client = apsecretclient.APSecretClient(self.key_file,
                                                       self.cert_file,
                                                       self.APSECRET_ENDPOINTS)

    # Determine whether to use "int" or "prod" for the MDSD publishing environment based on AP Environment
    # or on the parameter passed manually to script
    def determine_environment(self) -> str:
        try:
            current_environment = apenv.get_environment_name().lower()
        except Exception:
            raise ValueError("Could not obtain environment from apenv.get_environment_name()")

        if current_environment in self.mdsd_environment:
            mdsd_publish_environment = self.mdsd_environment[current_environment]
        else:
            mdsd_publish_environment = self.default_mdsd_environment

        return mdsd_publish_environment

    # Decode a base64 string
    def decode_base64_string(self, base64string: str) -> bytes:
        decoded_string = base64.b64decode(base64string, validate=True)

        return decoded_string

    # Go through autokey dictionary of known Datacenters and determine which region should be used for GCS
    def match_gcs_region(self) -> str:
        # Get the current machine region from autopilot.ini
        machine_region = apenv.get_cluster_name().lower()

        # See whether the region of the machine exists in the default mappings at the top of this file
        # If not, assign to default region
        if machine_region.upper() in self.gcs_regions:
            current_gcs_region = self.gcs_regions[machine_region.upper()]
        else:
            current_gcs_region = self.gcs_default_region

        if not current_gcs_region:
            raise ValueError("GCS for '{0}' Region does not exist".format(current_gcs_region))
        else:
            return current_gcs_region

    # Write string to a file; used to write secrets to chosen location
    def write_string_to_file(self, directory: str, file_path: str, string_to_write: str):
        if not os.path.isdir(directory):
            raise ValueError("Download directory '{0}' does not exist".format(directory))
        else:
            try:
                with open(file_path, "w", newline="\n") as text_file:
                    text_file.write(string_to_write)

            except Exception as e:
                print("apsecret_refresh_geneva_warm.write_string_to_file: Exception: " + str(e), file=sys.stderr)

    # Set permissions on written files so that the secrets are protected
    def set_permissions_on_file(self, full_file_path: str):
        # Set secret file mode to 0600 so that no one can mess with this but root
        try:
            os.chown(full_file_path, 0, 0)
            os.chmod(full_file_path, stat.S_IRUSR | stat.S_IWUSR)

        except Exception as e:
            print("apsecret_refresh_geneva_warm.set_permissions_on_file: Exception: " + str(e), file=sys.stderr)
            raise ValueError("Failed to set permissions on '{0}'".format(full_file_path))

    # Get the right Geneva certificates to use depending on environment
    def download_geneva_certificates(self, directory: str, environment: str):
        # Determine the Environment to use for certificates and folders based on which MDSD we are deploying
        self.__connect()

        # Determine the Environment to use for certificates for MDSD
        # If environment is passed in as variable honor it, if not, match it to the dictionary
        # defined in the __init__ function
        if not environment:
            environment = self.determine_environment()

        # Match GCS Region
        self.mdsd_gcs_region = self.match_gcs_region()

        if (environment == "prod"):
            self.mdsd_cert = self.kv_client.read_cache_secret(self.SECRET_MDSD_CERT)
            self.mdsd_key = self.kv_client.read_cache_secret(self.SECRET_MDSD_KEY)
            self.mdsd_gcs_environment = self.GCS_ENVIRONMENT
        elif (environment == "int"):
            self.mdsd_cert = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_CERT)
            self.mdsd_key = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_KEY)
            self.mdsd_gcs_environment = self.GCS_ENVIRONMENT_INT
        else:
            raise ValueError("Environment not specified and it is required")

        # Decode MDSD Cert from base64
        decoded_mdsd_cert = self.decode_base64_string(self.mdsd_cert).decode('utf-8')
        # Decode MDSD Key from base64
        decoded_mdsd_key = self.decode_base64_string(self.mdsd_key).decode('utf-8')

        # Make sure the download directory exists. If it does not, then we were
        # just refreshing the secret cache
        if not os.path.isdir(directory):
            return

        # Write MDSD .env file for GCS Environment, GCS Region and MDSD Account Name
        mdsd_gcs_file_path = os.path.join(directory, self.GCS_ENV_FILENAME)
        mdsd_export_string = 'export MONITORING_GCS_ENVIRONMENT="' + self.mdsd_gcs_environment + '"'
        mdsd_export_string += "\n" + 'export MONITORING_GCS_ACCOUNT="' + self.GCS_ACCOUNT + '"'
        mdsd_export_string += "\n" + 'export MONITORING_GCS_REGION="' + self.mdsd_gcs_region + '"'
        self.write_string_to_file(directory, mdsd_gcs_file_path, mdsd_export_string)

        # Write Cert and Key Files
        mdsd_cert_file_path = os.path.join(directory, self.GCS_CERT_FILENAME)
        self.write_string_to_file(directory, mdsd_cert_file_path, decoded_mdsd_cert)
        self.set_permissions_on_file(mdsd_cert_file_path)

        mdsd_key_file_path = os.path.join(directory, self.GCS_KEY_FILENAME)
        self.write_string_to_file(directory, mdsd_key_file_path, decoded_mdsd_key)
        self.set_permissions_on_file(mdsd_key_file_path)


def main(geneva_data_directory: str, environment: str):
    client_key, client_cert = apenv.get_machine_function_cert()
    refresher = GenevaSecretRefresher(client_key, client_cert)

    try:
        refresher.download_geneva_certificates(geneva_data_directory, environment)

    except Exception as e:
        print("ERROR: {0}".format(e), file=sys.stderr)
        sys.exit(1)

    sys.exit(0)


# Example run command:
# /ap/app/aptools/apsecret_refresh_geneva_warm.py -d /ap/data/geneva/mdsd_azuresecuritypack -e prod -n azsecpack
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Refreshes secrets for accessing Geneva Warm Path and copies "
                                     "certificates to the specified directory.")
    parser.add_argument("-d", "--directory", metavar="DIRECTORY", help="Directory to copy certificates into")
    parser.add_argument("-e", "--environment", metavar="ENVIRONMENT", help="Environment to use - prod/int")

    args = parser.parse_args()

    if args.directory is None or not os.path.exists(args.directory):
        print("ERROR: Valid Download directory must be specified.", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    main(args.directory, args.environment)
